<!doctype html>
<html lang="en">
<head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <meta name="generator" content="pdoc {{ __version__ }}"/>
    <title>{% block title %}{% endblock %}</title>
    {% block favicon %}
        {% if favicon %}
            <link rel="icon" href="{{ favicon }}"/>
        {% endif %}
    {% endblock %}
    {% block head %}{% endblock %}
    {% filter minify_css | indent %}
        {% block style %}
            <style>{% include "resources/bootstrap-reboot.min.css" %}</style>
            <style>/*! syntax-highlighting.css */{% include "syntax-highlighting.css" %}</style>
            {#
            The style_pdoc, style_theme, style_layout, and style_content Jinja2 blocks are deprecated and will be
            removed in a future release. Custom templates should either provide alternatives for the specific CSS files,
            or append their own styles by providing `custom.css` (see examples/custom-template/).
            #}
            {% block style_pdoc %}
                {% block style_theme %}<style>/*! theme.css */{% include "theme.css" %}</style>{% endblock %}
                {% block style_layout %}<style>/*! layout.css */{% include "layout.css" %}</style>{% endblock %}
                {% block style_content %}<style>/*! content.css */{% include "content.css" %}</style>{% endblock %}
                {# Use this file in your custom template directory to add additional CSS styling: #}
                <style>/*! custom.css */{% include "custom.css" %}</style>
            {% endblock %}
        {% endblock %}
    {% endfilter %}
    {% if math %}{% include "math.html.jinja2" %}{% endif %}
</head>
<body>
{% block body %}
    <nav class="pdoc">
        <label id="navtoggle" for="togglestate" class="pdoc-button">{% include 'resources/navtoggle.svg' %}</label>
        <input id="togglestate" type="checkbox" aria-hidden="true" tabindex="-1">
        <div>{% block nav %}{% endblock %}</div>
    </nav>
    {% block content %}{% endblock %}
{% endblock body %}
</body>
</html>
