/* pdoc color scheme (see also: examples/dark-mode) */
:root {
    --pdoc-background: #fff;
}

.pdoc {
    --text: #212529;
    --muted: #6c757d;
    --link: #3660a5;
    --link-hover: #1659c5;
    --code: #f8f8f8;
    --active: #fff598;

    --accent: #eee;
    --accent2: #c1c1c1;

    --nav-hover: rgba(255, 255, 255, 0.5);
    --name: #0066BB;
    --def: #008800;
    --annotation: #007020;
}
